# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###

import glob
import os
import sys


__all__ = ('available_commands', 'load_cli')


def available_commands():
    current_directory = os.path.dirname(__file__)
    for path in glob.glob('{}/*.py'.format(current_directory)):
        filename = os.path.basename(path)
        command_name = filename[:-3]
        if not command_name.startswith('_'):
            yield command_name


def _import_attr_n_module(module_name, attr):
    """From the given ``module_name`` import
    the value for ``attr`` (attribute).
    """
    __import__(module_name)
    module = sys.modules[module_name]
    attr = getattr(module, attr)
    return attr, module


def _import_loader(module_name):
    """Given a ``module_name`` import the cli loader."""
    loader, module = _import_attr_n_module(module_name, 'cli_loader')
    return loader, module.__doc__


def load_cli(subparsers):
    """Given a parser, load the CLI subcommands"""
    for command_name in available_commands():
        module = '{}.{}'.format(__package__, command_name)
        loader, description = _import_loader(module)
        parser = subparsers.add_parser(command_name,
                                       description=description)
        command = loader(parser)
        if command is None:
            raise RuntimeError('Failed to load "{}".'.format(command_name))
        parser.set_defaults(cmmd=command)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2016, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###

from contextlib import contextmanager
from io import StringIO
import os.path
import sys

import pip


here = os.path.abspath(os.path.dirname(__file__))
db_connection_string = 'dbname=travis user=travis host=localhost'
test_data_path = os.path.join(here, 'data')
test_packages = ['package-a', 'package-b']
test_config_path = os.path.join(test_data_path, 'config.ini')
test_migrations_directories = [
    os.path.join(test_data_path, 'package-a', 'package_a', 'migrations'),
    os.path.join(test_data_path, 'package-b', 'package_b', 'm'),
    ]


# noqa from http://stackoverflow.com/questions/4219717/how-to-assert-output-with-nosetest-unittest-in-python
@contextmanager
def captured_output():
    if sys.version_info[0] == 3:
        new_out, new_err = StringIO(), StringIO()
    else:
        from io import BytesIO
        new_out, new_err = BytesIO(), BytesIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


def install_test_packages(packages=None):
    if packages is None:
        packages = test_packages
    for package in packages:
        with captured_output() as (out, err):
            pip.main(['install', '-e', os.path.join(test_data_path, package)])
        stderr = err.getvalue()
        sys.stderr.write(stderr)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
from __future__ import print_function
import argparse
import logging
import os
import signal
import sys

import pkg_resources
import psycopg2

from . import commands, utils, logger


DEFAULTS = {
    }


def main(argv=sys.argv[1:]):
    # psycopg2 / libpq doesn't respond to SIGINT (ctrl-c):
    # https://github.com/psycopg/psycopg2/issues/333
    # To get around this problem, using code from:
    # http://initd.org/psycopg/articles/2014/07/20/cancelling-postgresql-statements-python/ # noqa
    psycopg2.extensions.set_wait_callback(utils.wait_select_inter)

    parser = argparse.ArgumentParser(description='DB Migrator')

    parser.add_argument('--verbose', '-v', action='store_true')

    parser.add_argument('--migrations-directory',
                        action='append',
                        default=[])

    parser.add_argument('--config')

    parser.add_argument('--db-connection-string',
                        help='a psycopg2 db connection string')

    parser.add_argument(
        '--super-user',
        default='postgres',
        help='postgres username for super user connections, defaults to '
             '"postgres"')

    parser.add_argument(
        '--context',
        action='append',
        default=[],
        help='Name of the python package containing the migrations')

    version = pkg_resources.require('db-migrator')[0].version
    parser.add_argument('-V', action='version', version=version,
                        help='Show version information')

    subparsers = parser.add_subparsers(help='commands')
    commands.load_cli(subparsers)

    args = parser.parse_args(argv)
    args = vars(args)

    if args.get('config'):
        if not os.path.exists(args['config']):
            raise Exception('config file not found')
        utils.get_settings_from_config(args['config'], [
            'migrations-directory',
            'db-connection-string',
            ], args)

    if not args.get('context') and not args.get('migrations_directory'):
        args['context'] = [os.path.basename(os.path.abspath(os.path.curdir))]
        print('context undefined, using current directory name "{}"'
              .format(args['context']),
              file=sys.stderr)
    utils.get_settings_from_entry_points(args, args['context'])

    for name, value in DEFAULTS.items():
        if not args.get(name):
            args[name] = value

    if 'cmmd' not in args:
        parser.print_help()
        return parser.error('command missing')

    if args.get('migrations_directory'):
        if not isinstance(args['migrations_directory'], list):
            args['migrations_directory'] = [args['migrations_directory']]
        args['migrations_directory'] = [
            os.path.relpath(md) for md in args['migrations_directory']]
    else:
        print('migrations directory undefined', file=sys.stderr)

    if args.get('verbose'):
        logger.setLevel(logging.DEBUG)

    logger.debug('args: {}'.format(args))
    utils.set_settings(args)

    return args['cmmd'](**args)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2016, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###

import os
import shutil
import sys
import tempfile
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

import pkg_resources
import psycopg2

from . import testing


class BaseTestCase(unittest.TestCase):
    @property
    def target(self):
        from ..cli import main
        return main

    def tearDown(self):
        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute('DROP TABLE IF EXISTS schema_migrations')


class VersionTestCase(BaseTestCase):
    def test(self):
        version = pkg_resources.get_distribution('db-migrator').version
        with testing.captured_output() as (out, err):
            self.assertRaises(SystemExit, self.target, ['-V'])

        stdout = out.getvalue()
        stderr = err.getvalue()
        if sys.version_info[0] == 3:
            self.assertEqual(stdout.strip(), version)
        else:
            self.assertEqual(stdout, '')
            self.assertEqual(stderr.strip(), version)


class VerboseTestCase(BaseTestCase):
    @mock.patch('dbmigrator.logger.debug')
    def test(self, debug):
        from .. import logger

        with testing.captured_output() as (out, err):
            self.target(['-v', '--config', testing.test_config_path, 'init'])

        stdout = out.getvalue()
        stderr = err.getvalue()

        self.assertEqual(1, debug.call_count)
        self.assertTrue(debug.call_args[0][0].startswith('args: {'))

        self.assertEqual('Schema migrations initialized.\n', stdout)
        self.assertEqual('', stderr)


class ListTestCase(BaseTestCase):
    def test_no_migrations_directory(self):
        cmd = ['--db-connection-string', testing.db_connection_string]
        self.target(cmd + ['init'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['list'])

        stdout = out.getvalue()
        stderr = err.getvalue()

        self.assertEqual(stdout, """\
version        | name            | is applied | date applied
----------------------------------------------------------------------\n""")
        self.assertEqual(stderr, """\
context undefined, using current directory name "['db-migrator']"
migrations directory undefined\n""")

    @mock.patch('dbmigrator.logger.warning')
    def test_no_table(self, warning):
        cmd = ['--config', testing.test_config_path]
        with testing.captured_output() as (out, err):
            self.target(cmd + ['list'])

        stdout = out.getvalue()
        stderr = err.getvalue()

        self.assertEqual("""\
version        | name            | is applied | date applied
----------------------------------------------------------------------\n""",
                         stdout)

        warning.assert_called_with(
            'You may need to create the schema_migrations table: '
            'dbmigrator init --help')
        self.assertEqual('', stderr)

    def test(self):
        testing.install_test_packages()

        cmd = ['--db-connection-string', testing.db_connection_string]
        self.target(cmd + ['init'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['-v', '--context', 'package-a', 'list'])

        stdout = out.getvalue()
        stderr = err.getvalue()

        self.assertEqual("""\
version        | name            | is applied | date applied
----------------------------------------------------------------------
20160228202637   add_table         False        \

20160228212456   cool_stuff        False        \
\n""", stdout)
        self.assertEqual('', stderr)

    def test_wide(self):
        testing.install_test_packages()

        cmd = ['--db-connection-string', testing.db_connection_string]
        self.target(cmd + ['init'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['-v', '--context', 'package-a',
                               'list', '--wide'])

        stdout = out.getvalue()
        stderr = err.getvalue()

        self.assertEqual("""\
version        | name       | is applied | date applied
----------------------------------------------------------------------
20160228202637   add_table    False        \

20160228212456   cool_stuff   False        \
\n""", stdout)
        self.assertEqual('', stderr)


class InitTestCase(BaseTestCase):
    def test_multiple_contexts(self):
        testing.install_test_packages()

        cmd = ['--db-connection-string', testing.db_connection_string]
        self.target(cmd + ['--context', 'package-a', '--context', 'package-b',
                           'init'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', '--context',
                               'package-b', 'list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         True', stdout)
        self.assertIn('20160228210326   initial_data      True', stdout)
        self.assertIn('20160228212456   cool_stuff        True', stdout)


class MarkTestCase(BaseTestCase):
    def test_missing_t_f_d_option(self):
        cmd = ['--db-connection-string', testing.db_connection_string]

        self.target(cmd + ['--context', 'package-a', 'init'])

        with self.assertRaises(Exception) as cm:
            self.target(cmd + ['--context', 'package-a', 'mark',
                               '20160228212456'])

        self.assertEqual('-t, -f or -d must be supplied.', str(cm.exception))

    def test_mark_as_true(self):
        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string]

        self.target(cmd + ['--context', 'package-a', 'init', '--version', '0'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a',
                               'mark', '-t', '20160228212456'])

        stdout = out.getvalue()
        self.assertEqual('Migration 20160228212456 marked as completed\n',
                         stdout)

        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         False', stdout)
        self.assertIn('20160228212456   cool_stuff        True', stdout)

    def test_mark_as_true_already_true(self):
        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string]

        self.target(cmd + ['--context', 'package-a', 'init'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a',
                               'mark', '-t', '20160228212456'])

        stdout = out.getvalue()
        self.assertEqual('Migration 20160228212456 marked as completed\n',
                         stdout)

        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         True', stdout)
        self.assertIn('20160228212456   cool_stuff        True', stdout)

    def test_migration_not_found(self):
        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string]

        self.target(cmd + ['--context', 'package-a', 'init', '--version', '0'])

        with self.assertRaises(SystemExit) as cm:
            self.target(cmd + ['mark', '-t', '012345'])
        self.assertEqual('Migration 012345 not found', str(cm.exception))

        with self.assertRaises(SystemExit) as cm:
            self.target(cmd + ['mark', '-f', '012345'])
        self.assertEqual('Migration 012345 not found', str(cm.exception))

    def test_mark_as_false(self):
        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string,
               '--context', 'package-a']

        self.target(cmd + ['init'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['mark', '-f', '20160228202637'])

        stdout = out.getvalue()
        self.assertIn('Migration 20160228202637 marked as not been run',
                      stdout)

        with testing.captured_output() as (out, err):
            self.target(cmd + ['list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         False', stdout)
        self.assertIn('20160228212456   cool_stuff        True', stdout)

    def test_mark_as_false_already_false(self):
        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string]

        self.target(cmd + ['--context', 'package-a', 'init', '--version', '0'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'mark', '-f',
                               '20160228202637'])

        stdout = out.getvalue()
        self.assertIn('Migration 20160228202637 marked as not been run',
                      stdout)

        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         False', stdout)
        self.assertIn('20160228212456   cool_stuff        False', stdout)

    def test_mark_as_deferred(self):
        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string]

        self.target(cmd + ['--context', 'package-a', 'init', '--version', '0'])

        # mark migration as deferred
        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'mark', '-d',
                               '20160228202637'])

        stdout = out.getvalue()
        self.assertEqual('Migration 20160228202637 marked as deferred\n',
                         stdout)

        # check list output
        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         deferred', stdout)
        self.assertIn('20160228212456   cool_stuff        False', stdout)

        # try running migrations
        self.target(cmd + ['--context', 'package-a', 'migrate'])

        # check that the table is not created by the migration
        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute("""\
SELECT 1 FROM information_schema.tables
    WHERE table_schema = 'public' AND table_name = 'a_table'""")
                table_exists = cursor.fetchone()
        self.assertEqual(None, table_exists)

        # check list output again
        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         deferred', stdout)
        self.assertIn('20160228212456   cool_stuff        True', stdout)

        # check that rollback does not rollback a deferred migration
        self.target(cmd + ['--context', 'package-a', 'rollback', '--step=2'])

        # check list output again
        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'list'])

        stdout = out.getvalue()
        self.assertIn('20160228202637   add_table         deferred', stdout)
        self.assertIn('20160228212456   cool_stuff        False', stdout)

    def test_deferred(self):
        md = os.path.join(testing.test_data_path, 'md')
        cmd = ['--db-connection-string', testing.db_connection_string,
               '--migrations-directory', md]

        self.target(cmd + ['init', '--version', '0'])

        # check list output
        with testing.captured_output() as (out, err):
            self.target(cmd + ['list'])

        stdout = out.getvalue()
        self.assertIn('20170810124056   empty             deferred', stdout)

        # mark a deferred migration as completed
        with testing.captured_output() as (out, err):
            self.target(cmd + ['mark', '-t', '20170810124056'])

        stdout = out.getvalue()
        self.assertEqual('Migration 20170810124056 marked as completed\n',
                         stdout)

        with testing.captured_output() as (out, err):
            self.target(cmd + ['list'])

        stdout = out.getvalue()
        self.assertIn('20170810124056   empty             deferred*    20',
                      stdout)

        # mark a deferred migration as not not been run
        with testing.captured_output() as (out, err):
            self.target(cmd + ['mark', '-f', '20170810124056'])

        stdout = out.getvalue()
        self.assertIn('Migration 20170810124056 marked as not been run',
                      stdout)

        with testing.captured_output() as (out, err):
            self.target(cmd + ['list'])

        stdout = out.getvalue()
        self.assertIn('20170810124056   empty             deferred', stdout)


class GenerateTestCase(BaseTestCase):
    @mock.patch('dbmigrator.utils.timestamp')
    def test(self, timestamp):
        timestamp.return_value = '20160423231932'
        filename = '{}_a_new_migration.py'.format(timestamp())
        expected_path = os.path.join(
            testing.test_migrations_directories[0], filename)

        def cleanup():
            if os.path.exists(expected_path):
                os.remove(expected_path)
        self.addCleanup(cleanup)

        testing.install_test_packages()

        with testing.captured_output() as (out, err):
            self.target(['--context', 'package-a', 'generate',
                         'a_new_migration'])

        stdout = out.getvalue()
        stderr = err.getvalue()

        self.assertEqual(
            'Generated migration script "dbmigrator/tests/data/package-a/'
            'package_a/migrations/{}"\n'.format(filename),
            stdout)
        self.assertEqual('', stderr)

        self.assertTrue(os.path.exists(expected_path))

        with open(expected_path, 'r') as f:
            content = f.read()

        self.assertIn('# -*- coding: utf-8 -*-', content)
        self.assertIn('def up(cursor):', content)
        self.assertIn('def down(cursor):', content)

    def test_no_migrations_directory(self):
        with self.assertRaises(Exception) as cm:
            self.target(['generate', 'a_new_migration'])

        self.assertEqual('migrations directory undefined',
                         str(cm.exception))

    def test_multiple_migrations_directory(self):
        tmp_dir = tempfile.gettempdir()
        tmp_dir2 = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmp_dir2)
        with self.assertRaises(Exception) as cm:
            self.target(['--migrations-directory', tmp_dir,
                         '--migrations-directory', tmp_dir2,
                         'generate', 'a_new_migration'])

        self.assertEqual('more than one migrations directory specified',
                         str(cm.exception))

    @mock.patch('dbmigrator.utils.timestamp')
    def test_migrations_directory_does_not_exist(self, timestamp):
        timestamp.return_value = '20160423231932'
        filename = '{}_a_new_migration.py'.format(timestamp())
        tmp_dir = tempfile.gettempdir()
        directory = '{}/dbmigrator-tests/m'.format(tmp_dir)
        expected_path = os.path.join(directory, filename)

        self.addCleanup(shutil.rmtree, directory)

        self.target(['--migrations-directory', directory,
                     'generate', 'a_new_migration'])

        self.assertTrue(os.path.exists(expected_path))


class MigrateTestCase(BaseTestCase):
    def test(self):
        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string]

        def cleanup():
            with psycopg2.connect(testing.db_connection_string) as db_conn:
                with db_conn.cursor() as cursor:
                    cursor.execute('DROP TABLE IF EXISTS a_table')

        self.addCleanup(cleanup)

        self.target(cmd + ['--context', 'package-a', 'init', '--version', '0'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['--context', 'package-a', 'migrate'])

        stdout = out.getvalue()
        self.assertIn('+CREATE TABLE a_table', stdout)

        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute("""\
                    SELECT table_name FROM information_schema.tables
                        WHERE table_name = 'a_table'""")
                self.assertEqual([('a_table',)], cursor.fetchall())

    def test_repeat(self):
        md = os.path.join(testing.test_data_path, 'md')
        cmd = ['--db-connection-string', testing.db_connection_string,
               '--migrations-directory', md]

        def cleanup():
            if os.path.exists('insert_data.txt'):
                os.remove('insert_data.txt')
            with psycopg2.connect(testing.db_connection_string) as db_conn:
                with db_conn.cursor() as cursor:
                    cursor.execute('DROP TABLE IF EXISTS a_table')

        self.addCleanup(cleanup)

        self.target(cmd + ['init', '--version', '0'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        stdout = out.getvalue()
        self.assertIn('+CREATE TABLE a_table', stdout)
        self.assertIn('Skipping migration 20170810093943', stdout)

        # Run the repeat migration by creating this file
        with open('insert_data.txt', 'w') as f:
            f.write('三好')

        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        stdout = out.getvalue()
        self.assertIn('Running migration 20170810093943', stdout)

        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        stdout = out.getvalue()
        self.assertIn('Running migration 20170810093943', stdout)

        # Make sure data has been inserted twice
        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute('SELECT name FROM a_table')
                self.assertEqual([('三好',), ('三好',)],
                                 cursor.fetchall())

        # Mark the repeat migration as deferred
        self.target(cmd + ['mark', '-d', '20170810093943'])

        # The deferred repeat migration should not run
        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        stdout = out.getvalue()
        self.assertIn('Skipping deferred migration 20170810124056 empty',
                      stdout)

    def test_deferred(self):
        md = os.path.join(testing.test_data_path, 'md')
        cmd = ['--db-connection-string', testing.db_connection_string,
               '--migrations-directory', md]

        def cleanup():
            if os.path.exists('insert_data.txt'):
                os.remove('insert_data.txt')
            with psycopg2.connect(testing.db_connection_string) as db_conn:
                with db_conn.cursor() as cursor:
                    cursor.execute('DROP TABLE IF EXISTS a_table')

        self.addCleanup(cleanup)

        self.target(cmd + ['init', '--version', '0'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        stdout = out.getvalue()
        self.assertIn('+CREATE TABLE a_table', stdout)
        self.assertIn('Skipping deferred migration 20170810124056 empty',
                      stdout)

        # Run the repeat migration by creating this file
        with open('insert_data.txt', 'w') as f:
            f.write('三好')

        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        stdout = out.getvalue()
        self.assertIn('Running migration 20170810093943', stdout)

        # Mark the deferred migration as not not been run
        self.target(cmd + ['mark', '-f', '20170810124056'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate', '--run-deferred'])
        stdout = out.getvalue()
        self.assertIn('Running migration 20170810124056', stdout)

        # Mark the deferable as deferred
        self.target(cmd + ['mark', '-d', '20170810124056'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate', '--run-deferred'])

        stdout = out.getvalue()
        self.assertIn('Running migration 20170810124056', stdout)


class RollbackTestCase(BaseTestCase):
    @mock.patch('dbmigrator.utils.timestamp')
    def test(self, timestamp):
        timestamp.return_value = '20160423231932'

        testing.install_test_packages()
        cmd = ['--db-connection-string', testing.db_connection_string,
               '--context', 'package-a']

        def cleanup():
            with psycopg2.connect(testing.db_connection_string) as db_conn:
                with db_conn.cursor() as cursor:
                    cursor.execute('DROP TABLE IF EXISTS a_table')
            path = os.path.join(
                testing.test_migrations_directories[0],
                '{}_new_one.py'.format(timestamp()))
            if os.path.exists(path):
                os.remove(path)

        self.addCleanup(cleanup)

        self.target(cmd + ['init', '--version', '0'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        # Generate a new migration that is not yet run
        self.target(cmd + ['generate', 'new_one'])

        # Rollback once, the table should still be there
        with testing.captured_output() as (out, err):
            self.target(cmd + ['rollback'])
        stdout = out.getvalue()
        self.assertIn('Rolling back migration 20160228212456', stdout)

        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute("""\
SELECT table_name FROM information_schema.tables
    WHERE table_name = 'a_table'""")
                self.assertEqual('a_table', cursor.fetchone()[0])

        # Rollback the migration that created the table
        with testing.captured_output() as (out, err):
            self.target(cmd + ['rollback'])
        stdout = out.getvalue()
        self.assertIn('Rolling back migration 20160228202637', stdout)

        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute("""\
SELECT table_name FROM information_schema.tables
    WHERE table_name = 'a_table'""")
                self.assertEqual(None, cursor.fetchone())

        # Migrate again
        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        # Rollback three migrations
        with testing.captured_output() as (out, err):
            self.target(cmd + ['rollback', '--steps', '3'])
        stdout = out.getvalue()
        self.assertIn('Rolling back migration 20160228212456', stdout)
        self.assertIn('Rolling back migration 20160228202637', stdout)

        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute("""\
SELECT table_name FROM information_schema.tables
    WHERE table_name = 'a_table'""")
                self.assertEqual(None, cursor.fetchone())

    def test_repeat(self):
        md = os.path.join(testing.test_data_path, 'md')
        cmd = ['--db-connection-string', testing.db_connection_string,
               '--migrations-directory', md]

        def cleanup():
            if os.path.exists('insert_data.txt'):
                os.remove('insert_data.txt')
            with psycopg2.connect(testing.db_connection_string) as db_conn:
                with db_conn.cursor() as cursor:
                    cursor.execute('DROP TABLE IF EXISTS a_table')

        self.addCleanup(cleanup)

        self.target(cmd + ['init', '--version', '0'])
        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate', '--run-deferred'])

        # Run the repeat migration by creating this file
        with open('insert_data.txt', 'w') as f:
            f.write('三好')

        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        with open('insert_data.txt', 'w') as f:
            f.write('ジョーカーゲーム')

        with testing.captured_output() as (out, err):
            self.target(cmd + ['migrate'])

        with testing.captured_output() as (out, err):
            self.target(cmd + ['list'])

        stdout = out.getvalue()
        self.assertIn('20170810093842   create_a_table    True ', stdout)
        self.assertIn('20170810093943   repeat_insert_d   True*', stdout)
        self.assertIn('20170810124056   empty             deferred*', stdout)

        # Version wise, the empty migration is more recent, but the repeat
        # migration is the last migration applied, so rollback should rollback
        # the repeat migration.
        with testing.captured_output() as (out, err):
            self.target(cmd + ['rollback'])
        stdout = out.getvalue()
        self.assertIn('Rolling back migration 20170810093943', stdout)

        with psycopg2.connect(testing.db_connection_string) as db_conn:
            with db_conn.cursor() as cursor:
                cursor.execute('SELECT name FROM a_table')
                self.assertEqual([('三好',)], cursor.fetchall())

        # Next, rollback the empty migration
        with testing.captured_output() as (out, err):
            self.target(cmd + ['rollback'])
        stdout = out.getvalue()
        self.assertIn('Rolling back migration 20170810124056', stdout)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2016, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
"""List migration versions, names, whether it has been applied and the date
applied."""

from .. import utils


__all__ = ('cli_loader',)


@utils.with_cursor
def cli_command(cursor, migrations_directory='', db_connection_string='',
                wide=False, **kwargs):
    # version -> applied timestamp
    migrated_versions = dict(list(
        utils.get_schema_versions(cursor, versions_only=False,
                                  raise_error=False)))
    migrations = utils.get_migrations(migrations_directory,
                                      import_modules=True)

    if wide:
        migrations = list(migrations)
        name_width = max([len(name) for _, name, _ in migrations])
    else:
        name_width = 15

    name_format = '{:<%s}' % (name_width,)
    print('version        | {} | is applied | date applied'
          .format(name_format.format('name')))
    print('-' * 70)

    name_format = '{: <%s}' % (name_width,)
    for version, migration_name, migration in migrations:
        applied_timestamp = migrated_versions.get(version, '')
        deferred = utils.is_deferred(
            version, migration, migrated_versions)
        is_applied = deferred and 'deferred' or \
            bool(migrated_versions.get(version))
        if hasattr(migration, 'should_run'):
            is_applied = '{}*'.format(is_applied)
        print('{}   {}   {!s: <10}   {}'.format(
            version, name_format.format(migration_name[:name_width]),
            is_applied, migrated_versions.get(version) or ''))


def cli_loader(parser):
    parser.add_argument('--wide', action='store_true',
                        dest='wide', default=False,
                        help='Display the full name of every migration')
    return cli_command



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2015, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
"""Run all pending migrations."""

from .. import utils


__all__ = ('cli_loader',)


@utils.with_cursor
def cli_command(cursor, migrations_directory='', version='',
                db_connection_string='', run_deferred=False, **kwargs):
    pending_migrations = utils.get_pending_migrations(
        migrations_directory, cursor, import_modules=True,
        up_to_version=version, include_defers=True)

    migrated = False
    for version, migration_name, migration in pending_migrations:
        migrated = True
        utils.compare_schema(db_connection_string,
                             utils.run_migration,
                             cursor,
                             version,
                             migration_name,
                             migration,
                             run_deferred)

    if not migrated:
        print('No pending migrations.  Database is up to date.')


def cli_loader(parser):
    parser.add_argument('--version',
                        help='Migrate database up to this version')
    parser.add_argument('--run-deferred',
                        action='store_true',
                        help='Also run the deferred migrations')
    return cli_command
